#include "Atlantis.h"
#include "level_data.h"
#include "fxlib.h"
#include <stdlib.h>
#include <stdio.h>

// 16 * 11 bitmaps
static const unsigned char boats[8][22] = {
	// Up.
	{ 241, 255, 192, 127, 192, 127, 192, 127, 192, 127, 128, 63, 224, 255, 128, 63, 241, 255, 192, 127, 251, 255 },
	{ 14, 0, 49, 128, 32, 128, 32, 128, 46, 128, 117, 192, 21, 0, 113, 192, 10, 0, 59, 128, 4, 0 },
	// Right.
	{ 192, 255, 192, 127, 224, 63, 224, 63, 224, 63, 192, 127, 192, 255, 59, 63, 0, 63, 128, 127, 192, 255 },
	{ 63, 0, 49, 128, 24, 192, 24, 192, 24, 192, 49, 128, 63, 0, 196, 192, 255, 192, 64, 128, 63, 0 },
	// Down.
	{ 251, 255, 192, 127, 224, 255, 128, 63, 192, 127, 128, 63, 192, 127, 192, 127, 224, 255, 224, 255, 241, 255 },
	{ 4, 0, 59, 128, 17, 0, 117, 192, 46, 128, 96, 192, 32, 128, 49, 128, 31, 0, 21, 0, 14, 0 },
	// Left.
	{ 192, 255, 128, 255, 1, 255, 1, 255, 1, 255, 128, 255, 192, 255, 55, 63, 0, 63, 128, 127, 192, 255 },
	{ 63, 0, 99, 0, 198, 0, 198, 0, 198, 0, 99, 0, 63, 0, 200, 192, 255, 192, 64, 128, 63, 0 }
};

/*
	Stats. Number after '+' indicate the increase applied for each level.

			PV	Atk	Def
	Medusa		15 +3	4  +1	0 +1
	Squid		25 +5	9  +2	3 +2
	Snake		20 +4	7  +1	2 +2
	Swordfish	35 +7	12 +3	3 +1
	Frigate		20 +6	6  +2	1 +1
	Destroyer	40 +10	10 +2	4 +2
	FAC		15 +3	8  +3	1 +1
*/

const int stats[][6] = {
	{ 15, 3,  4,  1, 0, 1 },
	{ 25, 5,  9,  2, 3, 2 },
	{ 20, 4,  7,  1, 2, 2 },
	{ 35, 7,  12, 3, 3, 1 },
	{ 20, 6,  6,  2, 1, 1 },
	{ 40, 10, 10, 2, 4, 2 },
	{ 15, 3,  8,  3, 1, 1 }
};

/*
	Functions and file-scope variables.
*/

static void draw_map(int x, int y, struct Map map, struct Entity *entities,
	struct Entity player);
static void draw_panel(int x, int y, struct Map map, struct Entity *entities,
	struct Entity player, int mode);

static int get_cell(int x, int y, struct Map map, struct Entity *entities);
static struct Entity *get_entity(int x, int y, struct Entity *entities);

static struct Entity *copy_entities(struct Map map);
static void set_entities(struct Entity *entities);

static void attack(struct Entity *entity, struct Entity *target);
static void end_turn(struct Map map, struct Entity *entities,
	struct Entity *players);

/*
	Function implementation.
*/

void level_enter(int place)
{
	struct Map map = maps[0];
	struct Entity *entities, *e;
	struct Entity player;
	int player_moves = data.turn_moves;
	int player_attacks = data.turn_attacks;

	unsigned int key;
	int cursor_x, cursor_y;
	int cell;
	int game_ended = 0;

	// Configuring the various entities.
	entities = copy_entities(map);
	set_entities(entities);

	player.identity = 0;
	player.level = data.level;

	player.x = map.initial_x;
	player.y = map.initial_y;
	player.direction = map.initial_direction;

	player.life = data.life;
	player.atk = data.atk;
	player.def = data.def;

	cursor_x = player.x;
	cursor_y = player.y;

	dclear();
	Bdisp_DrawLineVRAM(43, 0, 43, 63);
	SaveDisp(1);

	while(!game_ended)
	{
		RestoreDisp(1);
		draw_map(cursor_x, cursor_y, map, entities, player);
		draw_panel(cursor_x, cursor_y, map, entities, player, 0);

		GetKey(&key);

		switch(key)
		{
		case KEY_CTRL_UP:
			if(cursor_y > 0) cursor_y--;
			break;
		case KEY_CTRL_RIGHT:
			if(cursor_x < map.width - 1) cursor_x++;
			break;
		case KEY_CTRL_DOWN:
			if(cursor_y < map.height - 1) cursor_y++;
			break;
		case KEY_CTRL_LEFT:
			if(cursor_x > 0) cursor_x--;
			break;

		case KEY_CTRL_F1:
			if(!player_attacks) break;

			if(len(cursor_x, cursor_y, player.x, player.y) != 1)
				break;
			e = get_entity(cursor_x, cursor_y, entities);
			if(!e) break;
			if(e->identity <= ENT_ENEMY && e->life)
			{
				attack(&player, e);
				player_attacks--;
			}
			break;			

		case KEY_CTRL_ALPHA:
			if(!player_moves) break;

			if(len(cursor_x, cursor_y, player.x, player.y) != 1)
				break;
			cell = get_cell(cursor_x, cursor_y, map, entities);
			if(cell == -1)
			{
				e = get_entity(cursor_x, cursor_y, entities);
				if(e && e->identity == ENT_LEVELEND)
				{
					PopUpWin(1);
					locate(3, 4);
					Print((unsigned char *)
						"Vous avez gagne");
					game_ended = 1;
					break;
				}
			}

			if(cell != 0) break;

			player.x = cursor_x;
			player.y = cursor_y;
			player_moves--;
			break;
		}

		if((!player_moves && !player_attacks) || key == KEY_CTRL_EXE)
		{
			end_turn(map, entities, &player);
			if(!player.life)
			{
				PopUpWin(1);
				locate(3, 4);
				Print((unsigned char *)"Vous avez perdu");
				GetKey(&key);
				game_ended = 1;
			}

			player_moves = data.turn_moves;
			player_attacks = data.turn_attacks;
		}

	}

	free(entities);
}



/*
	Drawing functions.
*/

static void draw_map(int x, int y, struct Map map, struct Entity *entities,
	struct Entity player)
{
	const char *data = map.terrain;
	int ox = 44, oy = 2;
	int a = x - 3, b = y - 2;
	int i, j;
	int cell;

	// Determining the position of the top-left corner.
	if(a < 0) a = 0;
	else if(a + 6 >= map.width) a = map.width - 7;
	if(b < 0) b = 0;
	else if(b + 4 >= map.height) b = map.height - 5;

	// Drawing at ox:oy.
	for(i = 0; i < 8; i++) for(j = -1; j < 6; j++)
	{
		cell = get_cell(a + i, b + j, map, NULL);
		if(!cell || cell == -1) continue;

		dbitmap(tileset[cell - 1], ox + 12 * i, oy + 12 * j, 12, 12,
			MODE_OR);
	}

	// Drawing entities.
	for(i = 0; entities[i].identity; i++)
	{
		struct Entity e = entities[i];
		int px = ox + 12 * (e.x - a);
		int py = oy + 12 * (e.y - b);

		if(e.identity <= ENT_ENEMY && !e.life) continue;
		if(e.x < a || e.x >= a + 8) continue;
		if(e.y < b || e.y >= b + 6) continue;
		dbitmap(sprites[e.identity - 1], px, py, 16, 12, MODE_OR);
	}

	if(player.x >= a && player.x < a + 8 && player.y >= b
		&& player.y < b + 6)
	{
		dbitmap(boats[player.direction],
			ox + 12 * (player.x - a), oy + 12 * (player.y - b),
			16, 11, MODE_AND);
		dbitmap(boats[player.direction + 1],
			ox + 12 * (player.x - a), oy + 12 * (player.y - b),
			16, 11, MODE_OR);
	}

	// Drawing cursor.
	ox += 12 * (x - a);
	oy += 12 * (y - b);
	Bdisp_DrawLineVRAM(ox, oy, ox + 11, oy);
	Bdisp_DrawLineVRAM(ox, oy, ox, oy + 11);
	Bdisp_DrawLineVRAM(ox + 11, oy, ox + 11, oy + 11);
	Bdisp_DrawLineVRAM(ox, oy + 11, ox + 11, oy + 11);
}

static void draw_panel(int x, int y, struct Map map, struct Entity *entities,
	struct Entity player, int mode)
{
	#define	stat(str, val, y) \
		PrintMini(4, y, (const unsigned char *)str, 0); \
		sprintf(ch, "%d", val); \
		PrintMini(27, y, (const unsigned char *)ch, 0);

	char ch[10];

	struct Entity *e = get_entity(x, y, entities);
	if(!e && player.x == x && player.y == y) e = &player;

	if(e)
	{
		if(e == &player)
		{
			dbitmap(boats[1], 2, 9, 12, 11, MODE_OR);
			PrintMini(2, 2, (const unsigned char *)"Atlantis", 0);
		}
		else
		{
			dbitmap(sprites[e->identity - 1], 2, 8, 12, 12,
				MODE_OR);
			PrintMini(2, 2, spritenames[e->identity - 1], 0);
		}


		if(e == &player || e->identity <= ENT_ENEMY)
		{
			stat("Life", e->life, 24);
			stat("Atk", e->atk, 31);
			stat("Def", e->def, 38);
		}
	}
	else
	{
		int cell = get_cell(x, y, map, NULL);
		if(!cell || cell == -1) return;

		dbitmap(tileset[cell - 1], 2, 8, 12, 12, MODE_OR);
		PrintMini(2, 2, tilenames[cell - 1], 0);
	}

	#undef	stat
}


/*
	Board analysis.
*/

static int get_cell(int x, int y, struct Map map, struct Entity *entities)
{
	int offset = y * map.width + x;
	if(x < 0 || x >= map.width || y < 0 || y >= map.height) return -1;

	if(!entities || map.terrain[offset]) return map.terrain[offset];

	while(entities->identity)
	{
		if(entities->x == x && entities->y == y)
		{
			if(entities->identity > ENT_ENEMY || entities->life)
				return -1;
		}
		entities++;
	}
	return map.terrain[offset];
}

static struct Entity *get_entity(int x, int y, struct Entity *entities)
{
	while(entities->identity)
	{
		if(entities->x == x && entities->y == y)
		{
			if(entities->identity > ENT_ENEMY || entities->life)
				return entities;
		}
		entities++;
	}
	return NULL;
}



/*
	Entity manipulation.
*/

/*
	copy_entities()
	Copies entity data from the constant map data to allow editing.
*/

static struct Entity *copy_entities(struct Map map)
{
	struct Entity *entities;
	int i = 0, n;
	while(map.entities[i].identity) i++;

	entities = calloc(i, sizeof(*entities));
	for(n = 0; n <= i; n++) entities[n] = map.entities[n];
	return entities;
}

/*
	set_entities()
	Computes entities stats in compliance with their level.
*/

static void set_entities(struct Entity *entities)
{
	struct Entity *e = entities;
	const int *s;

	while(e->identity)
	{
		if(e->identity > ENT_ENEMY)
		{
			e++;
			continue;
		}

		s = stats[e->identity - 1];
		e->life = s[0] + (e->level - 1) * s[1];
		e->atk  = s[2] + (e->level - 1) * s[3];
		e->def  = s[4] + (e->level - 1) * s[5];
		e++;
	}
}



/*
	Game functionnalities.
*/

static void attack(struct Entity *entity, struct Entity *target)
{
	if(target->def > entity->atk) return;
	target->life -= entity->atk - target->def;
	if(target->life < 0) target->life = 0;
}

static void end_turn(struct Map map, struct Entity *entities, struct Entity
	*player)
{
	int distance;

	while(entities->identity)
	{
		if(entities->identity > ENT_ENEMY || !entities->life)
		{
			entities++;
			continue;
		}

		/*
			Enemy IA.
		*/

		// Trying to reach the player.
		distance = len(entities->x, entities->y, player->x, player->y);
		if(distance == 2)
		{
			int dx = (player->x - entities->x) / 2;
			int dy = (player->y - entities->y) / 2;

			if(dx || dy)
			{
				int cell = get_cell(entities->x + dx,
					entities->y + dy, map, entities);
				if(!cell) entities->x += dx, entities->y += dy;
			}
		}

		// Trying to attack the player.
		distance = len(entities->x, entities->y, player->x, player->y);
		if(distance == 1)
		{
			attack(entities, player);
		}

		entities++;
	}
}
